package org.bdware.sc.crdt.planning;

import java.util.*;

public class SharableNetworkPlanning {
    protected String[] nodeIds;
    protected int[] writers;
    protected int[] readers;
    protected long maxDelay;

    protected int bandwidthDownload;
    protected int bandwidthUpload;
    protected int dataSize;


    protected long wDelay;
    protected long rDelay;
    protected long w2rDelay;


    protected double totalCountW;
    protected double rootCountW;
    protected double treeNodeCountW;
    protected double treeDegreeW;
    protected double treeHeightW;
    protected double frequencySyncW;


    protected double totalCountR;
    protected double rootCountR;
    protected double treeNodeCountR;
    protected double treeDegreeR;
    protected double treeHeightR;
    protected double frequencySyncR;

    protected double frequencySyncWR;

    protected long totalData;

    public void adjustWriterTree(int rootCountW, int treeDegreeW) {
        this.rootCountW = rootCountW;
        this.treeDegreeW = treeDegreeW;
        this.treeNodeCountW = Math.ceil(totalCountW / rootCountW);
        this.treeHeightW = Math.ceil(logNM(treeDegreeW, treeNodeCountW * (treeDegreeW - 1) + 1));
    }

    public void adjustReaderTree(int rootCountR, int treeDegreeR) {
        this.rootCountR = rootCountR;
        this.treeDegreeR = treeDegreeR;
        this.treeNodeCountR = Math.ceil(totalCountR / rootCountR);
        this.treeHeightR = Math.ceil(logNM(treeDegreeR, treeNodeCountR * (treeDegreeR - 1) + 1));
    }

    protected double logNM(double n, double m) {
        return Math.log(m) / Math.log(n);
    }

    protected void adjustAndCalc() {
        long minTotalData = Long.MAX_VALUE;
        String result = "";
        for (int rootCountW = 1; rootCountW <= writers.length; ++rootCountW) {
            for (int treeDegreeW = 2; treeDegreeW <= writers.length / rootCountW - 1; ++treeDegreeW) {
                adjustWriterTree(rootCountW, treeDegreeW);
                for (int rootCountR = 1; rootCountR <= readers.length; ++rootCountR) {
                    for (int treeDegreeR = 2; treeDegreeR <= readers.length / rootCountR - 1; ++treeDegreeR) {
                        adjustReaderTree(rootCountR, treeDegreeR);
                        calcOptimizedResult();
                        if (!readerTreeConstraint()) {
                            //System.out.println("reader");
                            continue;
                        }
                        if (!writerTreeConstraint()) {
                            //System.out.println("writer");
                            continue;
                        }
                        if (!writer2ReaderConstraint()) {
                            //System.out.println("writer2Reader");
                            continue;
                        }
                        if (minTotalData > totalData) {
                            minTotalData = totalData;
                            result = toString();
                        }
                    }
                }
            }
        }

        System.out.println(minTotalData);
        System.out.println(result);
    }

    protected void calcOptimizedResult() {
    }

    protected boolean writer2ReaderConstraint() {
        return true;
    }

    protected boolean writerTreeConstraint() {
        return true;
    }

    protected boolean readerTreeConstraint() {
        return true;
    }

    public void allocate() {
        Set<Integer> writerOnlySet = new LinkedHashSet<>();
        for (int i : writers) {
            writerOnlySet.add(i);
        }
        Set<Integer> readerOnlySet = new LinkedHashSet<>();
        Set<Integer> rwSet = new LinkedHashSet<>();
        for (int i : readers) {
            if (writerOnlySet.contains(i)) {
                rwSet.add(i);
                writerOnlySet.remove(i);
            } else {
                readerOnlySet.add(i);
            }
        }
        int[] writerParents = allocateTreeNode(writerOnlySet, rwSet, (int) rootCountW, (int) treeDegreeW);
        int[] readerParents = allocateTreeNode(readerOnlySet, rwSet, (int) rootCountR, (int) treeDegreeR);
        System.out.println(readerParents);
    }

    private int[] allocateTreeNode(Set<Integer> onlySet, Set<Integer> rwSet, int rootCount, int degree) {
        int[] result = new int[nodeIds.length];
        Arrays.fill(result, -2);

        Map<Integer, List<Integer>> children = new HashMap<>();
        Deque<Integer> notFullNodesIdx = new LinkedList<>();
        for (Integer idx : onlySet) {
            if (children.size() < rootCount) {
                children.put(idx, new ArrayList<>());
                notFullNodesIdx.add(idx);
                result[idx] = -1;
            } else {
                int parentIdx = notFullNodesIdx.peek();
                children.computeIfAbsent(parentIdx, k -> new ArrayList<>()).add(idx);
                notFullNodesIdx.addLast(idx);
                if (children.get(parentIdx).size() >= degree) {
                    notFullNodesIdx.pop();
                }
            }
        }

        // 读写节点放在尽量靠近叶子节点，以便带宽的平衡
        for (Integer idx : rwSet) {
            if (children.size() < rootCount) {
                children.put(idx, new ArrayList<>());
                notFullNodesIdx.add(idx);
                result[idx] = -1;
            } else {
                int parentIdx = notFullNodesIdx.peek();
                children.computeIfAbsent(parentIdx, k -> new ArrayList<>()).add(idx);
                notFullNodesIdx.addLast(idx);
                if (children.get(parentIdx).size() >= degree) {
                    notFullNodesIdx.pop();
                }
            }
        }

        for (Integer parentId : children.keySet()) {
            for (int childId : children.get(parentId)) {
                result[childId] = parentId;
            }
        }
        return result;
    }

    @Override
    public String toString() {
        StringBuilder result = new StringBuilder();
        result.append("[").append(wDelay).append(",").append(w2rDelay).append(",").append(rDelay).append("]\n")
                .append("reader tree: degree ").append(treeDegreeR).append(", count ").append(rootCountR).append(",\n")
                .append("writer tree: degree ").append(treeDegreeW).append(", count ").append(rootCountW);
        return result.toString();
    }
}
